package org.jenkinsci.plugins.p4_client.client;

import hudson.security.ACL;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import jenkins.model.Jenkins;

import org.acegisecurity.Authentication;
import org.jenkinsci.plugins.p4_client.credentials.P4StandardCredentials;
import org.jenkinsci.plugins.p4_client.workspace.ManualWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.StaticWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.StreamWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.TemplateWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.Workspace;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.cloudbees.plugins.credentials.CredentialsProvider;
import com.cloudbees.plugins.credentials.domains.DomainRequirement;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.client.IClientSummary.ClientLineEnd;
import com.perforce.p4java.core.IMapEntry.EntryType;
import com.perforce.p4java.core.file.FileAction;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.impl.generic.client.ClientOptions;
import com.perforce.p4java.impl.generic.client.ClientView;
import com.perforce.p4java.impl.generic.client.ClientView.ClientViewMapping;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.option.client.ReconcileFilesOptions;
import com.perforce.p4java.option.client.RevertFilesOptions;
import com.perforce.p4java.option.client.SyncOptions;
import com.perforce.p4java.server.IOptionsServer;

public class ConnectionHelper {

	private Logger logger = LoggerFactory.getLogger(ConnectionHelper.class);

	private final ConnectionConfig connectionConfig;
	private final AuthorisationConfig authorisationConfig;
	private final IOptionsServer connection;
	private IClient iclient;

	public ConnectionHelper(String credentialID) throws Exception {
		P4StandardCredentials credential = findCredential(credentialID);
		this.connectionConfig = new ConnectionConfig(credential);
		this.authorisationConfig = new AuthorisationConfig(credential);
		this.connection = ConnectionFactory.getConnection(connectionConfig);
	}

	public ConnectionHelper(P4StandardCredentials credential) throws Exception {
		this.connectionConfig = new ConnectionConfig(credential);
		this.authorisationConfig = new AuthorisationConfig(credential);
		this.connection = ConnectionFactory.getConnection(connectionConfig);
	}

	public boolean isConnected() {
		return connection.isConnected();
	}

	public boolean isUnicode() {
		try {
			return connection.getServerInfo().isUnicodeEnabled();
		} catch (Exception e) {
			return false;
		}
	}

	public boolean login() throws Exception {
		connection.setUserName(authorisationConfig.getUsername());

		// CHARSET is not defined (only for client access)
		if (connection.getServerInfo().isUnicodeEnabled()) {
			connection.setCharsetName("utf8");
		}

		// exit early if logged in
		if (isLogin())
			return true;

		switch (authorisationConfig.getType()) {
		case PASSWORD:
			String pass = authorisationConfig.getPassword();
			connection.login(pass);
			break;

		case TICKET:
			String ticket = authorisationConfig.getTicketValue();
			connection.setAuthTicket(ticket);
			break;

		case TICKETPATH:
			String path = authorisationConfig.getTicketPath();
			connection.setTicketsFilePath(path);
			break;

		default:
			throw new Exception("Unknown Authorisation type: "
					+ authorisationConfig.getType());
		}

		return isLogin();
	}

	public void logout() throws Exception {
		if (isLogin()) {
			connection.logout();
		}
	}

	private boolean isLogin() throws Exception {
		String status = connection.getLoginStatus();
		if (status.contains("not necessary")) {
			return true;
		}
		if (status.contains("ticket expires in")) {
			return true;
		}
		// If there is a broker or something else that swallows the message
		if (status.isEmpty()) {
			return true;
		}

		logger.info("P4:login failed '" + status + "'");
		return false;
	}

	/**
	 * Create default client from Configuration
	 * 
	 * @param user
	 * @return
	 * @throws Exception
	 */
	public void setClient(Workspace workspace) throws Exception {

		if (connectionConfig.isUnicode()) {
			String charset = "utf8";
			connection.setCharsetName(charset);
		}

		switch (workspace.getType()) {
		case MANUAL:
			ManualWorkspaceImpl manualWs = (ManualWorkspaceImpl) workspace;
			iclient = connection.getClient(manualWs.getName());
			if (iclient == null) {
				logger.info("... Creating manual client: " + manualWs.getName());
				Client implClient = new Client();
				implClient.setName(manualWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(manualWs.getName());
			}

			ClientOptions options = new ClientOptions();
			options.setAllWrite(manualWs.allwrite);
			options.setClobber(manualWs.clobber);
			options.setCompress(manualWs.compress);
			options.setLocked(manualWs.locked);
			options.setModtime(manualWs.modtime);
			options.setRmdir(manualWs.rmdir);
			iclient.setOptions(options);

			iclient.setLineEnd(parseLineEnd(manualWs.getLine()));

			ClientView clientView = new ClientView();
			int order = 0;
			for (String line : manualWs.getView().split("\\n")) {
				ClientViewMapping entry = new ClientViewMapping(order, line);
				order++;
				clientView.addEntry(entry);
			}
			iclient.setClientView(clientView);

			break;

		case STATIC:
			StaticWorkspaceImpl staticWs = (StaticWorkspaceImpl) workspace;
			iclient = connection.getClient(staticWs.getName());
			if (iclient == null) {
				throw new Exception("Undefined workspace: "
						+ staticWs.getName());
			}
			break;

		case TEMPLATED:
			TemplateWorkspaceImpl templateWs = (TemplateWorkspaceImpl) workspace;
			String template = templateWs.getTemplateName();
			templateWs.getName();
			iclient = connection.getClient(templateWs.getName());
			if (iclient == null) {
				logger.info("Creating template client: " + templateWs.getName());
				Client implClient = new Client();
				implClient.setName(templateWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(templateWs.getName());
			}
			connection.switchClientView(template, templateWs.getName(), null);
			break;

		case STREAM:
			StreamWorkspaceImpl streamWs = (StreamWorkspaceImpl) workspace;
			if (iclient == null) {
				logger.info("Creating stream client: " + streamWs.getName());
				Client implClient = new Client();
				implClient.setName(streamWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(streamWs.getName());
			}
			iclient.setStream(streamWs.getStreamName());
			break;

		default:
			throw new Exception("Unknown workspace type: "
					+ workspace.getType());
		}

		// Ensure root and host fields are set
		iclient.setRoot(workspace.getRootPath());
		iclient.setHostName(workspace.getHostName());

		// Set clobber on to ensure workspace is always good
		ClientOptions options = new ClientOptions();
		options.setClobber(true);
		iclient.setOptions(options);

		// Save client spec
		iclient.update();

		// Set active client for this connection
		connection.setCurrentClient(iclient);
	}

	/**
	 * Sync latest revision of files to workspace.
	 * 
	 * @throws Exception
	 */
	public void syncFiles() throws Exception {
		// build file revision spec
		List<IFileSpec> syncFiles;
		syncFiles = FileSpecBuilder.makeFileSpecList("//...");

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		List<IFileSpec> syncMsg = iclient.sync(syncFiles, syncOpts);
		validateFileSpecs(syncMsg, "file(s) up-to-date.");
	}

	/**
	 * Sync files to workspace at the specified change.
	 * 
	 * @param change
	 * @throws Exception
	 */
	public void syncFiles(int change) throws Exception {
		// build file revision spec
		List<IFileSpec> files;
		files = FileSpecBuilder.makeFileSpecList("//...@" + change);

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		List<IFileSpec> syncMsg = iclient.sync(files, syncOpts);
		validateFileSpecs(syncMsg, "file(s) up-to-date.");
	}

	/**
	 * Cleans up the Perforce workspace after a previous build. Removes all
	 * pending and abandoned files (equivalent to 'p4 revert -w').
	 * 
	 * @throws Exception
	 */
	public void tidyWorkspace() throws Exception {
		// relies on workspace view for scope.
		List<IFileSpec> files;
		files = FileSpecBuilder.makeFileSpecList("//...");

		// revert all pending and shelved revisions
		RevertFilesOptions rOpts = new RevertFilesOptions();
		List<IFileSpec> list = iclient.revertFiles(files, rOpts);
		validateFileSpecs(list, "not opened on this client");

		// check for added files and remove...
		for (IFileSpec file : list) {
			if (file.getAction() == FileAction.ABANDONED) {
				String path = depotToLocal(file);
				if (path != null) {
					File unlink = new File(path);
					unlink.delete();
				}
			}
		}

		// check status - find all extra files
		ReconcileFilesOptions statusOpts = new ReconcileFilesOptions();
		statusOpts.setOutsideAdd(true);
		statusOpts.setNoUpdate(true);
		List<IFileSpec> extra = iclient.reconcileFiles(files, statusOpts);
		validateFileSpecs(extra, "- no file(s) to reconcile");

		// remove added files
		for (IFileSpec e : extra) {
			String path = depotToLocal(e);
			if (path != null) {
				File unlink = new File(path);
				unlink.delete();
			}
		}

		// check status - find all missing or changed
		statusOpts = new ReconcileFilesOptions();
		statusOpts.setNoUpdate(true);
		List<IFileSpec> update = iclient.reconcileFiles(files, statusOpts);

		// force sync to update files
		if (validateFileSpecs(update)) {
			SyncOptions syncOpts = new SyncOptions();
			syncOpts.setForceUpdate(true);
			iclient.sync(update, syncOpts);
		}
	}

	/**
	 * Workaround for p4java bug. The 'setLocalSyntax(true)' option does not
	 * provide local syntax, so I have to use 'p4 where' to translate through
	 * the client view.
	 * 
	 * @param fileSpec
	 * @return
	 * @throws Exception
	 */
	private String depotToLocal(IFileSpec fileSpec) throws Exception {
		String depotPath = fileSpec.getDepotPathString();
		if (depotPath == null)
			return null;

		List<IFileSpec> dSpec = FileSpecBuilder.makeFileSpecList(depotPath);
		List<IFileSpec> lSpec = iclient.where(dSpec);
		String path = lSpec.get(0).getLocalPathString();
		return path;
	}

	/**
	 * Unshelve review into workspace. Workspace is sync'ed to head first then
	 * review unshelved.
	 * 
	 * @param review
	 * @throws Exception
	 */
	public void unshelveFiles(int review) throws Exception {
		// build file revision spec
		List<IFileSpec> files;
		files = FileSpecBuilder.makeFileSpecList("//...");

		// Sync workspace to head
		SyncOptions syncOpts = new SyncOptions();
		List<IFileSpec> syncMsg = iclient.sync(files, syncOpts);
		validateFileSpecs(syncMsg, "file(s) up-to-date.");

		// Unshelve change for review
		List<IFileSpec> shelveMsg;
		shelveMsg = iclient.unshelveChangelist(review, files, 0, true, false);
		validateFileSpecs(shelveMsg);

		// Remove opened files from have list.
		RevertFilesOptions rOpts = new RevertFilesOptions();
		rOpts.setNoUpdate(true);
		List<IFileSpec> rvtMsg = iclient.revertFiles(files, rOpts);
		validateFileSpecs(rvtMsg);
	}

	/**
	 * Disconnect from the Perforce Server.
	 * 
	 * @throws Exception
	 */
	public void disconnect() throws Exception {
		connection.disconnect();
	}

	/**
	 * Look for a message in the returned FileSpec from operation.
	 * 
	 * @param fileSpecs
	 * @param ignore
	 * @return
	 * @throws ConverterException
	 */
	public boolean validateFileSpecs(List<IFileSpec> fileSpecs,
			String... ignore) throws Exception {
		for (IFileSpec fileSpec : fileSpecs) {
			if (fileSpec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();

				// superfluous p4java message
				boolean unknownMsg = true;
				ArrayList<String> ignoreList = new ArrayList<String>();
				ignoreList.addAll(Arrays.asList(ignore));
				for (String istring : ignoreList) {
					if (msg.contains(istring)) {
						// its a known message
						unknownMsg = false;
					}
				}

				// check and report unknown message
				if (unknownMsg) {
					logger.warn("p4java: " + msg);
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Finds a Perforce Credential based on the String id.
	 * 
	 * @return a P4StandardCredentials credential or null if not found.
	 */
	private static P4StandardCredentials findCredential(String id) {
		Class<P4StandardCredentials> type = P4StandardCredentials.class;
		Jenkins scope = Jenkins.getInstance();
		Authentication acl = ACL.SYSTEM;
		DomainRequirement domain = new DomainRequirement();

		List<P4StandardCredentials> list;
		list = CredentialsProvider.lookupCredentials(type, scope, acl, domain);

		for (P4StandardCredentials c : list) {
			if (c.getId().equals(id)) {
				return c;
			}
		}
		return null;
	}

	private ClientLineEnd parseLineEnd(String line) {
		for (ClientLineEnd end : ClientLineEnd.values()) {
			end.name().equalsIgnoreCase(line);
			return end;
		}
		return ClientLineEnd.LOCAL;
	}

	private ClientViewMapping parseViewEntry(String line) {
		ClientViewMapping entry = new ClientViewMapping();

		if (line.startsWith("-")) {
			entry.setType(EntryType.EXCLUDE);
		} else if (line.startsWith("+")) {
			entry.setType(EntryType.OVERLAY);
		} else {
			entry.setType(EntryType.INCLUDE);
		}

		String[] parts = line.split("\\s+");
		entry.setLeft(parts[0]);
		entry.setRight(parts[1]);

		return entry;
	}

}
